/*
* Copyright 2025 Yubico AB
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
 */

#ifndef YUBIHSM_SHELL_CMD_UTIL_H
#define YUBIHSM_SHELL_CMD_UTIL_H

EVP_PKEY *wrap_public_key(yh_session *session, yh_algorithm algorithm,
                          EVP_PKEY *public_key, uint16_t key_id);
X509_NAME *parse_subject_name(const char *orig_name);

#endif // YUBIHSM_SHELL_CMD_UTIL_H
